#ifdef USE_SDL
#include "SDL.h"
#include "SDL_image.h"
//#include "SDL_gfxPrimitives.h"
//#include "SDL_rotozoom.h"
#include "SDL_draw.h"
#else
#include <stdio.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#define Uint8 unsigned char
#endif

#include <stdbool.h>
#include "../vecterm/vecterm_communication.h"

#include "osint.h"
#include "e8910.h"
#include "vecx.h"

#define EMU_TIMER 20 /* the emulators heart beats at 20 milliseconds */

#ifdef USE_SDL
static SDL_Surface *screen = NULL;
static SDL_Surface *overlay_original = NULL;
static SDL_Surface *overlay = NULL;
#endif // USE_SDL

static long scl_factor;
static long offx;
static long offy;

framegen_data *frame;

#define SCREEN_H 1024
#define SCREEN_W 1024
int oldx, oldy, currentx,currenty;

void osint_render(void){
#ifdef USE_SDL
        SDL_FillRect(screen, NULL, 0);
#endif // USE_SDL
	
	int v;
	for(v = 0; v < vector_draw_cnt; v++){
		Uint8 c = vectors_draw[v].color * 256 / VECTREX_COLORS;
#ifdef NEVER
		aalineRGBA(screen,
				offx + vectors_draw[v].x0 / scl_factor,
				offy + vectors_draw[v].y0 / scl_factor,
				offx + vectors_draw[v].x1 / scl_factor,
				offy + vectors_draw[v].y1 / scl_factor,
				c, c, c, 0xff);
#else
		/*void (*Draw_Line)(SDL_Surface *super,
				  Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2,
				  Uint32 color);*/

#ifdef USE_SDL
		Draw_Line(screen,
				oldx=offx + vectors_draw[v].x0 / scl_factor,
				oldy=offy + vectors_draw[v].y0 / scl_factor,
				currentx=offx + vectors_draw[v].x1 / scl_factor,
				currenty=offy + vectors_draw[v].y1 / scl_factor,
			        (((c<<8)|c)<<8)|c);
#else
		oldx=(vectors_draw[v].x0 / scl_factor);
		oldy=(vectors_draw[v].y0 / scl_factor)+200;
		currentx=(vectors_draw[v].x1 / scl_factor);
		currenty=(vectors_draw[v].y1 / scl_factor)+200;
#endif // USE_SDL
		if (frame->line_count < frame->max_line_count) {
		  float *fptr = frame->ptr + 5*frame->line_count++;
#ifdef USE_SDL
		  fptr[0] = ((float)oldx/(float)SCREEN_W)*2.0;
		  fptr[1] = ((float)(oldy+50)/(float)SCREEN_H)*1.6; // +50 to move everything up a little
		  fptr[2] = ((float)currentx/(float)SCREEN_W)*2.0;
		  fptr[3] = ((float)(currenty+50)/(float)SCREEN_H)*1.6;
		  fptr[4] = 1.0;//(float)(z&15)/15.0;
#else
		  fptr[0] = ((float)oldx/(float)SCREEN_W)*1.10;
		  fptr[1] = ((float)oldy/(float)SCREEN_H)*0.8; // +50 to move everything up a little
		  fptr[2] = ((float)currentx/(float)SCREEN_W)*1.10;
		  fptr[3] = ((float)currenty/(float)SCREEN_H)*0.8;
		  fptr[4] = ((float)c)/255.0;//(float)(z&15)/15.0;
		  //fprintf(stderr, "%d\n",c);
#endif //USE_SDL
		}
		
#endif
	}
#ifdef USE_SDL
	if(overlay){
		SDL_Rect dest_rect = {offx, offy, 0, 0};
		SDL_BlitSurface(overlay, NULL, screen, &dest_rect);
	}
#endif // USE_SDL
	
	//XX: vecterm_wait_for_elapsed_frame();
	vecterm_new_frame(frame);
	frame->line_count = 0;
	
#ifdef USE_SDL
	SDL_Flip(screen);
#endif // USE_SDL
}

static char *romfilename = "rom.dat";
static char *cartfilename = NULL;

static void init(){
	FILE *f;
	if(!(f = fopen(romfilename, "rb"))){
		perror(romfilename);
		exit(EXIT_FAILURE);
	}
	if(fread(rom, 1, sizeof (rom), f) != sizeof (rom)){
		printf("Invalid rom length\n");
		exit(EXIT_FAILURE);
	}
	fclose(f);

	memset(cart, 0, sizeof (cart));
	if(cartfilename){
		FILE *f;
		if(!(f = fopen(cartfilename, "rb"))){
			perror(cartfilename);
			exit(EXIT_FAILURE);
		}
		fread(cart, 1, sizeof (cart), f);
		fclose(f);
	}
}

void resize(int width, int height){
        long sclx, scly;
	long screenx = width;
	long screeny = height;
#ifdef USE_SDL

	screen = SDL_SetVideoMode(screenx, screeny, 0, SDL_SWSURFACE | SDL_RESIZABLE);

	sclx = ALG_MAX_X / screen->w;
	scly = ALG_MAX_Y / screen->h;

	scl_factor = sclx > scly ? sclx : scly;

	offx = (screenx - ALG_MAX_X / scl_factor) / 2;
	offy = (screeny - ALG_MAX_Y / scl_factor) / 2;

#ifdef NEVER
	if(overlay_original){
		if(overlay)
			SDL_FreeSurface(overlay);
		double overlay_scale = ((double)ALG_MAX_X / (double)scl_factor) / (double)overlay_original->w;
		overlay = zoomSurface(overlay_original, overlay_scale, overlay_scale, 0);
	}
#endif
#else // USE_SDL
	sclx = ALG_MAX_X / SCREEN_W;
	scly = ALG_MAX_Y / SCREEN_H;

	scl_factor = sclx > scly ? sclx : scly;

	offx = (screenx - ALG_MAX_X / scl_factor) / 2;
	offy = (screeny - ALG_MAX_Y / scl_factor) / 2;	
#endif // USE_SDL
}

static void readevents(){
#ifdef USE_SDL
  SDL_Event e;
	while(SDL_PollEvent(&e)){
		switch(e.type){
			case SDL_QUIT:
				exit(EXIT_SUCCESS);
				break;
			case SDL_VIDEORESIZE:
				resize(e.resize.w, e.resize.h);
				break;
			case SDL_KEYDOWN:
				switch(e.key.keysym.sym){
					case SDLK_ESCAPE:
						exit(EXIT_SUCCESS);
					case SDLK_a:
						snd_regs[14] &= ~0x01;
						break;
					case SDLK_s:
						snd_regs[14] &= ~0x02;
						break;
					case SDLK_d:
						snd_regs[14] &= ~0x04;
						break;
					case SDLK_f:
						snd_regs[14] &= ~0x08;
						break;
					case SDLK_LEFT:
						alg_jch0 = 0x00;
						break;
					case SDLK_RIGHT:
						alg_jch0 = 0xff;
						break;
					case SDLK_UP:
						alg_jch1 = 0xff;
						break;
					case SDLK_DOWN:
						alg_jch1 = 0x00;
						break;
					default:
						break;
				}
				break;
			case SDL_KEYUP:
				switch(e.key.keysym.sym){
					case SDLK_a:
						snd_regs[14] |= 0x01;
						break;
					case SDLK_s:
						snd_regs[14] |= 0x02;
						break;
					case SDLK_d:
						snd_regs[14] |= 0x04;
						break;
					case SDLK_f:
						snd_regs[14] |= 0x08;
						break;
					case SDLK_LEFT:
						alg_jch0 = 0x80;
						break;
					case SDLK_RIGHT:
						alg_jch0 = 0x80;
						break;
					case SDLK_UP:
						alg_jch1 = 0x80;
						break;
					case SDLK_DOWN:
						alg_jch1 = 0x80;
						break;
					default:
						break;
				}
				break;
			default:
				break;
		}
	}
#else
	int tmp_btns, mx, my;
	tmp_btns = vecterm_get_buttons(); // READ ONCE PER FRAME

	// digital for now.  May switch to analog later
	mx = vecterm_get_joystick_analog_x_controller_0();
	my = vecterm_get_joystick_analog_y_controller_0();

	if (tmp_btns & (VECTERM_BUTTON_0_1 | VECTERM_BUTTON_1_1))
	  snd_regs[14] &= ~0x01;
        else
	  snd_regs[14] |= 0x01;

	if (tmp_btns & (VECTERM_BUTTON_0_2 | VECTERM_BUTTON_1_2)) 
	  snd_regs[14] &= ~0x02;
	else
	  snd_regs[14] |= 0x02;
	
	if (tmp_btns & (VECTERM_BUTTON_0_3 | VECTERM_BUTTON_1_3)) 
	  snd_regs[14] &= ~0x04;
	else
	  snd_regs[14] |= 0x04;

	if (tmp_btns & (VECTERM_BUTTON_0_4 | VECTERM_BUTTON_1_4)) 
          snd_regs[14] &= ~0x08;
	else
          snd_regs[14] |= 0x08;


	alg_jch0 = (((signed char)mx)+0x80)&0xff;
        alg_jch1 = (((signed char)my)+0x80)&0xff;
	
#endif // USE_SDL
}

void osint_emuloop(){
#ifdef USE_SDL
  Uint32 next_time = SDL_GetTicks() + EMU_TIMER;
#endif
  vecx_reset();
	for(;;){
		vecx_emu((VECTREX_MHZ / 1000) * EMU_TIMER);
		readevents();
#ifdef USE_SDL
		{
			Uint32 now = SDL_GetTicks();
			if(now < next_time)
				SDL_Delay(next_time - now);
			else
				next_time = now;
			next_time += EMU_TIMER;
		}
#endif
	}
}

void load_overlay(const char *filename){
#ifdef USE_SDL
  SDL_Surface *image;
	image = IMG_Load(filename);
	if(image){
		overlay_original = image;
	}else{
		fprintf(stderr, "IMG_Load: %s\n", IMG_GetError());
	}
#else
  (void)filename;
#endif
}

static framegen_data *init_framegen_data(void) {
  static framegen_data frame;
  frame.options.use_ay_regs           = false;
  frame.options.use_led               = false;
  frame.options.use_led_dimmed        = false;
  frame.options.use_points            = false;

  frame.options.screen_swap_x_y       = false;
  frame.options.screen_flip_x         = false;
  frame.options.screen_flip_y         = false;
  frame.options.aspect_ratio_adjust   = false;
  switch (vecterm_orientation()) {
  case 3:
    frame.options.screen_swap_x_y       = true;
    frame.options.screen_flip_x         = true;
    frame.options.aspect_ratio_adjust   = true;
    break;
  case 2:
    frame.options.screen_flip_x         = true;
    frame.options.screen_flip_y         = true;
    break;
  case 1:
    break;
  default:
    frame.options.screen_swap_x_y       = true;
    frame.options.screen_flip_y         = true;
    frame.options.aspect_ratio_adjust   = true;
    break;
  }

  frame.options.aspect_ratio          = 1.0;//3.0/5.0;
  //frame.options.aspect_ratio          = 4.0/5.0; // this hack is compensated for by a counter-hack where SCREEN_H is changed from 800 to 1023 :-/

  frame.options.use_opt_1             = true;
  frame.options.use_opt_2             = true;
  frame.options.use_opt_point_offset  = false;
  frame.options.use_opt_point_char_detect  = false;
  frame.options.continue_distance     = 1;
  frame.options.point_size            = 8;

  // frame.measurement_flags             = VECTERM_MEASUREMENT_FLAGS_BUTTONS  | VECTERM_MEASUREMENT_FLAGS_DIGITAL_X | VECTERM_MEASUREMENT_FLAGS_DIGITAL_Y;
  frame.measurement_flags             = VECTERM_MEASUREMENT_FLAGS_BUTTONS  | VECTERM_MEASUREMENT_FLAGS_ANALOG_X | VECTERM_MEASUREMENT_FLAGS_ANALOG_Y;

  static float tmp_mem[5*16*1024];
  frame.max_line_count = 16*1024;
  frame.ptr = tmp_mem;
  return (&frame);
}
    
int main(int argc, char *argv[]){
#ifdef USE_SDL
        if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO) < 0){
		fprintf(stderr, "Failed to initialize SDL: %s\n", SDL_GetError());
		exit(-1);
	}
#endif
	
	frame = init_framegen_data();
	if (!vecterm_open_connection(0)) {
	  fprintf(stderr, "error: opening vecterm connection failed\n");
	  exit(-1);
	}
	
	resize(330*3/2, 410*3/2);

	if(argc > 1)
		cartfilename = argv[1];
	if(argc > 2)
		load_overlay(argv[2]);

	init();

	e8910_init_sound();
	osint_emuloop();
	e8910_done_sound();

#ifdef USE_SDL
	SDL_Quit();
#endif
	
	return 0;
}

